# flake8: noqa

'''
Serializers for xdgconfig. Each serializer is a module exposing a
`loads` and `dumps` function. These functions behave like `json.loads` and
`json.dumps`.

The `dumps` function needs a keyword argument named `indent` which behaves
exactly like the `indent` keyword argument in the `json` module.
'''

import contextlib

from xdgconfig.serializers import _json as json
from xdgconfig.serializers import _configparser as ini

with contextlib.suppress(ImportError):
    from xdgconfig.serializers import _xml as xml

with contextlib.suppress(ImportError):
    from xdgconfig.serializers import _toml as toml

with contextlib.suppress(ImportError):
    from xdgconfig.serializers import _yaml as yaml



import typer

from xdgconfig import Config
from xdgconfig.cli_tools.common import HELP_TEXT


typer_config = typer.Typer()


@typer_config.callback(invoke_without_command=True)
def config(
    context: typer.Context,
    config_key: str, config_value:str,
    _global: bool = typer.Option(False, '--global'),
):
    config = context.default_map.get('config')
    infer_type = context.default_map.get('infer_type', True)

    if config is None:
        raise typer.Exit(1)

    config._cli_callback(
        config_key, config_value,
        _global=_global, infer_type=infer_type
    )


def make_ctx_settings(config, infer_type=True):
    return {
        'default_map': {
            'config': config,
            'infer_type': infer_type,
        }
    }



from typing import Any


class defaultdict(dict):
    _DEFAULTS = {}

    def __init__(self, *args, **kwargs):
        '''
        A defaultdict which recursively replace dicts inside with itself.
        Supports a `_DEFAULTS` class attribute to specify defaults for each key.

        :param parent: The parent key. Defaults to None
        :type parent: str, optional
        :param defaults: The defaults to assign to the `_DEFAULTS`
                         class attribute. Defaults to dict().
        :type defaults: dict, optional
        '''
        parent = kwargs.pop('_parent', None)
        defaults = kwargs.pop('_defaults', {})
        super().__init__(*args, **kwargs)
        self._parent = parent
        self._DEFAULTS = defaults

    def __getitem__(self, key: str) -> Any:
        if key not in self and '.' not in key:
            default = self._default(key)
            if isinstance(default, dict):
                if self._parent is not None:
                    parent = f'{self._parent}.{key}'
                else:
                    parent = key
                default = defaultdict(
                    default, _parent=parent,
                    _defaults=self._DEFAULTS,
                )
            self[key] = default
        if '.' in key:
            k, *rest = key.split('.')
            return self[k]['.'.join(rest)]
        return super().__getitem__(key)

    def _default(self, key: str) -> Any:
        '''
        Returns the default value for a given key.

        :param key: The key to get the default value of.
        :type key: str
        :return: The default value associated with the key
        :rtype: Any
        '''
        if self._parent:
            path = f'{self._parent}.{key}'
        else:
            path = key
        return self._DEFAULTS.get(
            path, defaultdict(
                _parent=path, _defaults=self._DEFAULTS,
            )
        )

    def flat(self, parent_key='', sep='.'):
        items = []
        for k, v in self.items():
            new_key = parent_key + sep + k if parent_key else k
            if isinstance(v, defaultdict):
                items.extend(v.flat(new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)



import os
import pathlib
import subprocess

from xdgconfig.exceptions import GitNotFound


class Git:
    def __init__(self, path: pathlib.Path):
        self.cwd: pathlib.Path = pathlib.Path.cwd().resolve()
        self.path = path
        try:
            subprocess.call(['git', '--version'])
        except subprocess.CalledProcessError:
            raise GitNotFound('git is not installed on this system')

    def add(self, glob: str) -> None:
        os.chdir(self.path)
        subprocess.call(['git', 'add', glob])
        os.chdir(self.cwd)

    def commit(self, message: str) -> None:
        os.chdir(self.path)
        subprocess.call(['git', 'commit', '-m', message])
        os.chdir(self.cwd)

    def push(self, branch: str = 'master', remote: str = 'origin') -> None:
        os.chdir(self.path)
        subprocess.call(['git', 'push', remote, branch])
        os.chdir(self.cwd)

    def has_remote(self, remote: str = 'origin') -> bool:
        os.chdir(self.path)
        out: str = subprocess.check_output(['git', 'remote'])
        os.chdir(self.cwd)
        return remote in out.decode('utf8')

    @property
    def status(self) -> str:
        os.chdir(self.path)
        out: str = subprocess.check_output(['git', 'status'])
        os.chdir(self.cwd)
        return out.decode('utf8')



import os
import pathlib

from xdgconfig.config import Config


class UnixConfig(Config):
    @property
    def _base_path(self):
        return pathlib.Path(
            os.getenv('XDG_CONFIG_HOME', '~/.config')
        ).expanduser()

    def _load(self):
        default_config_path = (
            pathlib.Path('/etc') / f'{self._app_name}.d' / self._config_name
        )
        data = {}

        if default_config_path.exists():
            with open(default_config_path, 'r') as fp:
                data.update(self._SERIALIZER.loads(fp.read()))
        data.update(super()._load())
        return data



import importlib.util


class Plugin:
    def __init__(self, config, name):
        self.name = name
        base_path = config._config_path.parent
        spec = importlib.util.spec_from_file_location(
            f'plugins.{name}',
            base_path / name / '__init__.py'
        )
        self.module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self.module)
        if hasattr(self.module, 'on_load'):
            func = getattr(self.module, 'on_load')
            if callable(func):
             func(config['plugins'][name])

    def __getattr__(self, attr):
        if attr in ('name', 'module'):
            return super().__getattribute__(attr)
        return getattr(self.module, attr)



import configparser
import io
import json

from mergedeep import merge

from xdgconfig.utils import path_to_dict


__all__ = [
    'loads', 'dumps'
]


def _to_string(value):
    if isinstance(value, bool):
        return 'yes' if value else 'no'
    if isinstance(value, bytes):
        return value.decode('utf8')
    if isinstance(value, (list, tuple, set, frozenset, dict)):
        return json.dumps(value)
    if value is None:
        return ''
    return str(value)

def loads(contents, **kw):  # noqa
    config = configparser.ConfigParser()
    config.read_string(contents)

    data = {}
    for section in config.sections():
        for option in config.options(section):
            merge(
                data,
                path_to_dict(
                    section, option,
                    config.get(section, option)
                )
            )
    return data


def dump_recurse(config, data, previous_section=''):
    for section, contents in data.items():
        if previous_section:
            section_name = f'{previous_section}.{section}'
        else:
            section_name = section
        if isinstance(contents, dict):
            config = dump_recurse(config, contents, section_name)
            config[section_name] = {}
            for option, value in contents.items():
                if isinstance(value, dict):
                    continue
                config[section_name][option] = _to_string(value)
    return config


def dumps(data, **kw):  # noqa
    fp = io.StringIO()
    config = configparser.ConfigParser()
    config = dump_recurse(config, data)
    config.write(fp)
    fp.seek(0)
    return fp.read()



import datetime
import os
import pathlib
from typing import Any

from xdgconfig.utils import cast, default_to_dict, dict_to_default
from xdgconfig.defaultdict import defaultdict
from xdgconfig.plugin import Plugin
from xdgconfig.git import Git
from xdgconfig.exceptions import GitNotFound


__all__ = ('Config', 'LocalConfig')


class ConfigMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        for argname, arg in zip(('app_name', 'config_name'), args):
            kwargs[argname] = arg
        instance_name = kwargs['app_name'] + '.' + kwargs['config_name']
        if instance_name not in cls._instances:
            cls._instances[instance_name] = super(
                ConfigMeta, cls
            ).__call__(**kwargs)
        return cls._instances[instance_name]


class Config(defaultdict, metaclass=ConfigMeta):
    '''An object representing a configuration for an application'''

    _SERIALIZER = None

    def __init__(
        self, app_name: str, config_name: str = 'config', *,
        autosave: bool = True
    ) -> None:
        '''
        :param app_name: The name of your app
        :type name: str
        :param config_name: The name of the config file, defaults to 'config'
        :type config_name: str
        :param autosave: Whether to autosave the config on mutation,
                            defaults to True
        :type autosave: bool, optional
        '''

        self._app_name = app_name
        self._config_name = config_name
        self._autosave = autosave
        self._parent = None
        self._local = None

        self.plugins = []

        for key, value in self._load().items():
            super().__setitem__(key, value)

    def __setitem__(self, key: str, value: Any):
        if isinstance(value, dict):
            value = defaultdict(value, _parent=key, _defaults=self._DEFAULTS)
        super().__setitem__(key, value)
        if self._autosave:
            self.save()

    @property
    def _base_path(self) -> pathlib.Path:
        '''
        Abstract method that returns the base path of the config directory

        :return: The path to the configuration directory.
        :rtype: pathlib.Path
        '''
        raise NotImplementedError()

    def _setup_local_config(self):
        self._local = LocalConfig(
            f'.{self._app_name}', self._config_name,
            autosave=self._autosave
        )
        self._local._SERIALIZER = self._SERIALIZER

    @property
    def is_git_repo(self):
        return (self.app_path / '.git').exists()

    @property
    def app_path(self) -> pathlib.Path:
        '''Returns the path to the app's config directory'''
        return self._base_path / self._app_name

    @property
    def config_path(self) -> pathlib.Path:
        '''Returns the path to the config file'''
        return self.app_path / self._config_name

    def save(self) -> None:
        '''
        Saves the config to a file.
        '''
        self.config_path.parent.mkdir(parents=True, exist_ok=True)

        with open(self.config_path, 'w') as fp:
            data = self._SERIALIZER.dumps(default_to_dict(self), indent=4)
            fp.write(data)

        if self.is_git_repo:
            self.commit_changes()

    def commit_changes(self, push=True) -> None:
        '''
        Commit the changes to the config to a repo,
        push if a remote is available

        :param push: Whether to push to a remote, if available. Defaults to True.
        '''
        date: str = datetime.datetime.now().strftime('%Y-%m-%d')
        message: str = f'[{date}] Committing changes.'
        try:
            git = Git(self.app_path)
        except GitNotFound:
            return

        git.add('.')
        git.commit(message)
        if git.has_remote('origin') and push:
            git.push()

    def load_plugins(self):
        names = [p.name for p in self.plugins]
        for plugin_name in (self._base_path / self._app_name).iterdir():
            if plugin_name not in names:
                self.plugins.append(Plugin(self, plugin_name))

    def _load(self) -> dict:
        '''
        Loads the config into memory

        :return: A dictionnary representing the loaded configuration.
        :rtype: dict
        '''
        try:
            with open(self.config_path, 'r') as fp:
                data = self._SERIALIZER.loads(fp.read())
        except FileNotFoundError:
            data = dict()

        # PROG_CONFIG_PATH environment variable can be used to point to
        # a configuration file that will take precedence over the user config.
        environ = os.getenv(f'{self._app_name.upper()}_CONFIG_PATH')
        if environ and os.path.exists(environ):
            with open(environ) as fp:
                data.update(self._SERIALIZER.loads(fp.read()))
        return dict_to_default(data)

    def _cli_callback(
        self, config_key: str, config_value: str,
        _global: bool = True, infer_type: bool = True,
    ) -> int:
        '''
        Callback for CLIs to set the proper data in this object

        :param config_key: The key to set on the config object
        :type config_key: str
        :param config_value: The value associated with that key
        :type config_value: str
        :param _global: Whether to use the global config scope, defaults to True
        :type _global: bool, optional
        :param infer_type: Whether to infer the type from the value string, defaults to True
        :type infer_type: bool, optional
        :return: The exit code (0=OK)
        :rtype: int
        '''

        if not _global:
            return self._local._cli_callback(config_key, config_value)  # noqa

        if infer_type:
            config_value = cast(config_value)
        self[config_key] = config_value
        return 0


class LocalConfig(Config):
    @property
    def _base_path(self):
        return pathlib.Path.cwd().resolve()

    def _setup_local_config(self):
        return



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xdgconfig', 'xdgconfig.cli_tools', 'xdgconfig.serializers']

package_data = \
{'': ['*']}

install_requires = \
['mergedeep>=1.3.4,<2.0.0']

setup_kwargs = {
    'name': 'xdgconfig',
    'version': '1.1.2',
    'description': 'Easy access to `~/.config`',
    'long_description': "# XDGConfig\n\n[![Discord](https://img.shields.io/discord/812645240611536906?label=&logo=discord&logoColor=ffffff&color=7389D8&labelColor=6A7EC2)](https://discord.gg/kDgyr9Uzwj)\n[![GitHub license](https://img.shields.io/github/license/Dogeek/xdgconfig.svg)](https://github.com/Dogeek/xdgconfig/blob/master/LICENSE)\n[![PyPI version shields.io](https://img.shields.io/pypi/v/xdgconfig.svg)](https://pypi.python.org/pypi/xdgconfig/)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/xdgconfig.svg)](https://pypi.python.org/pypi/xdgconfig/)\n\nEasy access to `~/.config`.\n\n\n## Installation\n\n### Using pip\n\nSimply run `pip3 install --upgrade xdgconfig`.\n\nBy default, `xdgconfig` only supports JSON as its serializer, but you can install support for\nother serializers by specifiying the format in square brackets, i.e. `pip3 install xdgconfig[xml]`.\nThe following are available:\n\n- `jsonc`: JSON, with comments\n- `ini`: INI files\n- `xml`: eXtensible Markup Language files\n- `toml`: Tom's Markup language files\n- `yaml`: YAML Ain't Markup Language files\n\nFurthermore there is an `all` recipe to install support for every markup supported,\nand you can combine them by using a `+` between 2 targets, i.e. `pip3 install xdgconfig[xml+toml]`\n\n### From source\n\nSimply clone this repo and run `python3 setup.py install`.\n\n## Features\n\n- `Config` objects use a shared single reference.\n- Serializing to many common formats, including JSON, XML, TOML, YAML, and INI\n- `dict`-like interface\n- Autosaving on mutation of the `Config` object.\n- Smart config loading, especially on Unix-based platforms\n  - looks in `/etc/prog/config`, then in `~/.config/prog/config`\n  - Supports setting a config file path in an environment variable named `PROG_CONFIG_PATH`\n- Accessing the config using dot notation (`config.key` for instance). See limitations for guidance.\n\n\n## Usage\n\n```python\nfrom xdgconfig import JsonConfig\n\n# Instanciate the JsonConfig object\n# If you'd rather use a different format, there also are config classes\n# for TOML, YAML, INI (configparser), and XML.\n# This will save your configuration under `~/.config/PROG/config\nconfig = JsonConfig('PROG', autosave=True)\n\nconfig['foo'] = 'bar'  # Save a value to the config\n\n# Access the value later on\nprint(config['foo'])\n\n# It behaves like a collections.defaultdict as well\nconfig['oof']['bar'] = 'baz'\n\n# Prints {'oof': {'bar': 'baz'}, 'foo': 'bar'}\nprint(config)\n\n```\n\n## Adding onto the library\n\n### Custom serializers\n\nYou can add custom serializers support by using a Mixin class, as well as\na serializer class which must have a `dumps` and a `loads` method, which will\nbe used to store and load data from the config file. The data is always\nrepresented as a python `dict` object, but you can serialize any data you want\ninside of it.\n\nLook at the following example for an implementation guide.\n\n```python\nfrom typing import Any, Dict\n\nfrom xdgconfig import Config\n\n\nclass MySerializer:\n    def dumps(data: Dict[str, Any]) -> str:\n        return '\\n'.join(f'{k}:{v}' for k, v in data.items())\n\n    def loads(contents: str) -> Dict[str, Any]:\n        return dict(s.split(':') for s in contents.split('\\n'))\n\n\nclass MySerializerMixin:\n    _SERIALIZER = MySerializer\n\n\nclass MyConfig(MySerializerMixin, Config):\n    ...\n\n```\n\n### Setting default values\n\nYou can set default values by creating a `Mixin` class with a `_DEFAULTS` class attribute, such as :\n\n```python\nfrom pathlib import Path\nfrom pprint import pprint\n\nfrom xdgconfig import JsonConfig\n\n\nclass DefaultConfig:\n    _DEFAULTS = {\n        'logger.level': 'info',\n        'logger.verbosity': 3,\n        'app.path': str(Path.cwd()),\n        'app.credentials.username': 'user',\n        'app.credentials.password': 'password',\n    }\n\n\nclass Config(DefaultConfig, JsonConfig):\n    ...\n\n\nconfig = Config('PROG', 'config.json')\npprint(config)\n# Prints the following dict :\n# {\n#     'logger': {\n#         'level': 'info',\n#         'verbosity': 3\n#     },\n#     'app': {\n#         'path': '$CWD',\n#         'credentials': {\n#             'username': 'user',\n#             'password': 'password'\n#         }\n#     }\n# }\n\n```\n\n\n## Known limitations\n\n- Using an `IniConfig` object prevents you from using periods (`.`) in key names, as they are separators for subdicts.\n- Methods and attributes of the `Config` object all start with a leading underscore (`_`), hence, using key names with the same convention is discouraged, as it could break the object due to the way dot (`.`) accessing works. The only exception is the `save` method, which doesn't start with a leading underscore.\n- There can only be one document per config file, and a config file is a dictionary.\n- Depending on the serializer used, some data types may or may not be available. You can circumvent that by using custom serializers.\n- Configuration files with comments will have their comments dropped when the configuration is saved.\n",
    'author': 'Dogeek',
    'author_email': 'Dogeek@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Dogeek/xdgconfig',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import argparse
from functools import partial

from xdgconfig import Config
from xdgconfig.cli_tools.common import HELP_TEXT


def config(
    subparsers: argparse.Action, config: Config,
    infer_type: bool = True,
) -> argparse.Action:
    '''
    Configuration function that returns a pregenerated argparse subparser
    to add to your CLIs using `xdgconfig`. The subparser has a callback
    attached to it, which you can call with the arguments from itself.
    Example :
        args = vars(parser.parse_args())
        callback = args.pop('callback', lambda *a, **kw: None)
        callback(**args)

    :param subparsers: The return value of argparse.add_subparsers
    :type subparsers: argparse._SubParsersAction
    :param config: The configuration object from xdgconfig.
    :type config: xdgconfig.Config
    :param infer_type: Whether to infer types from the value apparent
                       type, defaults to True
    :type infer_type: bool, optional
    :return: The modified argparse._SubParsersAction
    :rtype: argparse._SubParsersAction
    '''

    config_p = subparsers.add_parser(
        'config', help=HELP_TEXT,
    )

    config_p.add_argument(
        'config_key', type=str, help='Config key to edit'
    )
    config_p.add_argument(
        'config_value', type=str, help='Value to assign to the config key.'
    )
    config_p.add_argument(
        '--global', action='store_true', dest='_global',
        help='Whether to assign to value to the global configuration'
    )
    config_p.set_defaults(
        callback=partial(
            config._cli_callback, infer_type=infer_type,  # noqa
        ),
    )
    return subparsers



import os
import pathlib

from xdgconfig.config import Config


class WinConfig(Config):
    @property
    def _base_path(self):
        return pathlib.Path(
            os.getenv('XDG_CONFIG_HOME', '~/AppData/Roaming')
        ).expanduser()



from copy import deepcopy
import re
from typing import Any, Dict

from xdgconfig.defaultdict import defaultdict


def cast(value: str) -> Any:
    if value in ('yes', 'no', 'true', 'false'):
        return value in ('yes', 'true')
    if value.isdigit():
        return int(value)
    if re.search(r'\d+\.\d+', value):
        return float(value)
    if value[2:].isdigit():
        if value.startswith('0x'):
            return int(value, base=16)
        if value.startswith('0o'):
            return int(value, base=8)
        if value.startswith('0b'):
            return int(value, base=2)
    return value


def path_to_dict(path: str, option: str, value: str) -> Dict[str, Any]:
    parts = path.split('.')

    def pack(parts):
        if parts:
            return {parts[0]: pack(parts[1:])}
        return {option: cast(value)}

    return pack(parts)


def default_to_dict(data: defaultdict) -> dict:
    data_ = deepcopy(dict(data))
    for k, v in data_.items():
        if isinstance(v, defaultdict):
            data_[k] = dict(default_to_dict(v))
    return data_


def dict_to_default(data: dict) -> defaultdict:
    data_ = deepcopy(dict(data))
    for k, v in data_.items():
        if isinstance(v, dict):
            data_[k] = defaultdict(dict_to_default(v))
    return data_
